/* grecs - Gray's Extensible Configuration System -*- c -*-
   Copyright (C) 2007-2016 Sergey Poznyakoff

   Grecs is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   Grecs is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with Grecs. If not, see <http://www.gnu.org/licenses/>. */

#ifndef _GRECS_H
#define _GRECS_H

#include <grecs/types.h>
#include <grecs/version.h>
#include <grecs/doc.h>
#include <grecs/error.h>
#include <grecs/format.h>
#include <grecs/lex.h>
#include <grecs/list.h>
#include <grecs/locus.h>
#include <grecs/mem.h>
#include <grecs/node.h>
#include <grecs/parser.h>
#include <grecs/preproc.h>
#include <grecs/sockaddr.h>
#include <grecs/symtab.h>
#include <grecs/table.h>
#include <grecs/tree.h>
#include <grecs/txtacc.h>
#include <grecs/util.h>
#include <grecs/value.h>

#endif
